@@ -14,9 +14,7 @@ let () =
         match C.Pkg_config.get c with
           | None -> default
           | Some pc -> (
-              match
-                C.Pkg_config.query pc ~package:"flac"
-              with
+              match C.Pkg_config.query pc ~package:"flac" with
                 | None -> default
                 | Some deps -> deps)
       in
